use std::cell::RefCell;
use std::collections::HashMap;
use std::rc::Rc;
use std::sync::{Arc, RwLock};

use super::render::vulkan_context::VulkanContext;
use crate::core::input::InputManager;
use crate::core::scene::manager::SceneManager;
use crate::core::timer::Timer;
use crate::game::scenes::main_scene::MainScene;
use egui_winit_vulkano::{Gui, GuiConfig};
use user_event::UserEvent;
use vulkano::format::Format;
use vulkano::image::ImageUsage;
use vulkano::swapchain::PresentMode;
use vulkano_util::context::VulkanoContext;
use vulkano_util::window::{VulkanoWindows, WindowDescriptor};
use winit::application::ApplicationHandler;
use winit::event::WindowEvent;
use winit::event_loop::{ActiveEventLoop, EventLoopProxy};
use winit::window::WindowId;

use self::context::WindowContext;

pub mod context;
pub mod user_event;

pub const DEPTH_IMAGE_ID: usize = 0;

pub struct App {
    vulkan_context: Arc<VulkanContext>,
    vulkano_windows: Rc<RefCell<VulkanoWindows>>,
    gui: HashMap<WindowId, Rc<RefCell<Gui>>>,
    scene_manager: HashMap<WindowId, SceneManager>,
    input_manager: Arc<RwLock<InputManager>>,
    timer: Arc<RwLock<Timer>>,
    event_loop_proxy: EventLoopProxy<UserEvent>,

    // Context d'application partagé par fenêtre - architecture unifiée
    app_contexts: HashMap<WindowId, Rc<RefCell<WindowContext>>>,
}

impl App {
    pub fn new(
        vulkano_context: VulkanoContext,
        input_manager: InputManager,
        event_loop_proxy: EventLoopProxy<UserEvent>,
    ) -> Self {
        Self {
            vulkan_context: Arc::new(VulkanContext::new(vulkano_context)),
            vulkano_windows: Rc::new(RefCell::new(VulkanoWindows::default())),
            gui: HashMap::new(),
            input_manager: Arc::new(RwLock::new(input_manager)),
            scene_manager: HashMap::new(),
            timer: Arc::new(RwLock::new(Timer::new())),
            event_loop_proxy,
            app_contexts: HashMap::new(),
        }
    }
}

impl ApplicationHandler<UserEvent> for App {
    fn resumed(&mut self, event_loop: &ActiveEventLoop) {
        let mut vulkano_windows = self.vulkano_windows.borrow_mut();
        let window_id = vulkano_windows.create_window(
            event_loop,
            self.vulkan_context.vulkano_context(),
            &WindowDescriptor {
                title: "Rust ASH Test".to_string(),
                width: 800.0,
                height: 600.0,
                present_mode: PresentMode::Fifo,
                cursor_visible: false,
                cursor_locked: true,
                ..Default::default()
            },
            |_| {},
        );

        let renderer = vulkano_windows.get_renderer_mut(window_id).unwrap();
        renderer.add_additional_image_view(
            DEPTH_IMAGE_ID,
            Format::D16_UNORM,
            ImageUsage::DEPTH_STENCIL_ATTACHMENT,
        );

        let gui = {
            Gui::new(
                event_loop,
                renderer.surface(),
                renderer.graphics_queue(),
                renderer.swapchain_format(),
                GuiConfig {
                    is_overlay: true,
                    allow_srgb_render_target: true,
                    ..Default::default()
                },
            )
        };
        self.gui.insert(window_id, Rc::new(RefCell::new(gui)));

        let mut scene_manager = SceneManager::new();
        scene_manager.load_scene(Box::new(MainScene::default()));

        self.scene_manager.insert(window_id, scene_manager);

        let app_context = Rc::new(RefCell::new(WindowContext::new(
            self.vulkan_context.clone(),
            self.vulkano_windows.clone(),
            self.input_manager.clone(),
            self.timer.clone(),
            self.gui.get(&window_id).unwrap().clone(),
            self.event_loop_proxy.clone(),
            window_id,
        )));
        self.app_contexts.insert(window_id, app_context);
    }

    fn device_event(
        &mut self,
        _event_loop: &ActiveEventLoop,
        _device_id: winit::event::DeviceId,
        event: winit::event::DeviceEvent,
    ) {
        let mut input_manager = self.input_manager.write().unwrap();
        input_manager.process_device_event(&event);
    }

    fn window_event(&mut self, event_loop: &ActiveEventLoop, id: WindowId, event: WindowEvent) {
        {
            let gui = self.gui.get_mut(&id).unwrap();
            let mut gui = gui.borrow_mut();
            if !gui.update(&event) {
                let mut input_manager = self.input_manager.write().unwrap();
                input_manager.process_window_event(&event);
            }
        }

        match event {
            WindowEvent::CloseRequested => {
                tracing::debug!("The close button was pressed; stopping");
                event_loop.exit();
            }
            WindowEvent::Resized(_) | WindowEvent::ScaleFactorChanged { .. } => {
                let mut vulkano_windows = self.vulkano_windows.borrow_mut();
                vulkano_windows.get_renderer_mut(id).unwrap().resize();
            }
            WindowEvent::RedrawRequested => {
                let _frame_span = tracing::info_span!("frame").entered();

                {
                    let _input_span = tracing::debug_span!("input_update").entered();
                    let mut input_manager = self
                        .input_manager
                        .write()
                        .expect("Failed to lock input manager");
                    input_manager.update();
                }
                {
                    let _timer_span = tracing::debug_span!("timer_update").entered();
                    let mut timer = self.timer.write().expect("Failed to lock timer");
                    timer.update();
                }

                // Créer ou mettre à jour le contexte d'application
                let window_context = self.app_contexts.get(&id).unwrap().clone();
                let scene_manager = self.scene_manager.get_mut(&id).unwrap();

                // Utiliser le contexte partagé pour les scènes
                {
                    let mut context = window_context.borrow_mut();

                    {
                        let _scene_span =
                            tracing::info_span!("scene_loading_if_not_loaded").entered();
                        scene_manager
                            .load_scene_if_not_loaded(&mut context)
                            .unwrap();
                    }

                    if let Some(scene) = scene_manager.current_scene_mut() {
                        {
                            let _update_span = tracing::debug_span!("scene_update").entered();
                            scene.update(&mut context).unwrap();
                        }

                        let acquire_future = {
                            let _acquire_span = tracing::debug_span!("acquire_swapchain").entered();
                            context.with_renderer_mut(|renderer| {
                                renderer.acquire(None, |_| {}).unwrap()
                            })
                        };

                        let acquire_future = {
                            let _render_span = tracing::debug_span!("scene_render").entered();
                            scene.render(acquire_future, &mut context).unwrap()
                        };

                        {
                            let _present_span = tracing::debug_span!("present_frame").entered();
                            context.with_renderer_mut(|renderer| {
                                renderer.present(acquire_future, true);
                            });
                        }
                    } else {
                        tracing::warn!("No current scene found for update!");
                    }
                }

                {
                    let _gui_span = tracing::debug_span!("request_redraw").entered();
                    let mut window_context = window_context.borrow_mut();
                    window_context.with_renderer_mut(|renderer| {
                        renderer.window().request_redraw();
                    });
                }
            }
            _ => {}
        }
    }

    fn user_event(&mut self, event_loop: &ActiveEventLoop, event: UserEvent) {
        match event {
            UserEvent::CursorGrabMode(window_id, grab) => {
                let vulkano_windows = self.vulkano_windows.borrow();
                let window = vulkano_windows.get_window(window_id).unwrap();
                if let Err(e) = window.set_cursor_grab(grab) {
                    tracing::error!("Failed to set cursor grab: {}", e);
                }
            }
            UserEvent::CursorVisible(window_id, visible) => {
                let vulkano_windows = self.vulkano_windows.borrow();
                let window = vulkano_windows.get_window(window_id).unwrap();
                window.set_cursor_visible(visible);
            }
            UserEvent::ChangeScene(window_id, scene) => {
                if let Some(scene_manager) = self.scene_manager.get_mut(&window_id) {
                    scene_manager.load_scene(scene);
                }
            }
            UserEvent::ChangeResolution(window_id, width, height) => {
                let mut vulkano_windows = self.vulkano_windows.borrow_mut();
                let window = vulkano_windows.get_window(window_id).unwrap();
                let _ = window.request_inner_size(winit::dpi::LogicalSize::new(width, height));
                let renderer = vulkano_windows.get_renderer_mut(window_id).unwrap();
                renderer.resize();
                tracing::trace!(
                    "Resolution changed to {}x{} for window {:?}",
                    width,
                    height,
                    window_id
                );
            }
            UserEvent::Exit(window_id) => {
                tracing::trace!("Exit requested for window {:?}", window_id);
                event_loop.exit();
            }
        }
    }
}
