use std::{
    collections::HashMap,
    sync::{Arc, RwLock},
};

use winit::{
    event::{AxisId, ElementState, MouseButton},
    keyboard::PhysicalKey,
};

use super::{
    virtual_binding::VirtualBinding,
    virtual_state::{VirtualBindingState, VirtualInputState},
};

#[derive(Default)]
pub struct VirtualInput {
    // Global states
    states: HashMap<String, Arc<RwLock<VirtualInputState>>>,

    // Per kind of input states to keep complexity low during state updates
    states_by_key: HashMap<PhysicalKey, Vec<Arc<RwLock<VirtualInputState>>>>,
    mouse_move_states: Vec<Arc<RwLock<VirtualInputState>>>,
    mouse_wheel_states: Vec<Arc<RwLock<VirtualInputState>>>,
    mouse_button_states: HashMap<MouseButton, Vec<Arc<RwLock<VirtualInputState>>>>,
    axis_states: HashMap<AxisId, Vec<Arc<RwLock<VirtualInputState>>>>,
}

impl std::fmt::Debug for VirtualInput {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut debug = f.debug_struct("VirtualInput");

        for (name, state) in &self.states {
            let value = state.read().expect("Poisoned lock for debug").value;
            debug.field(name, &value);
        }

        debug.finish()
    }
}

impl VirtualInput {
    pub fn get_state(&self, value_name: &str) -> f32 {
        self.states
            .get(value_name)
            .map(|state| state.read().expect("Poisoned lock for get state").value)
            .unwrap_or(0.0)
    }

    pub fn add_bindings(&mut self, value_name: String, new_bindings: Vec<VirtualBinding>) {
        let state = self
            .states
            .entry(value_name)
            .or_insert(Arc::new(RwLock::new(VirtualInputState {
                value: 0.0,
                bindings: Vec::new(),
            })));

        for binding in &new_bindings {
            match binding {
                VirtualBinding::Keyboard(key, _) => {
                    self.states_by_key
                        .entry(*key)
                        .or_default()
                        .push(state.clone());
                }
                VirtualBinding::MouseX(_) | VirtualBinding::MouseY(_) => {
                    self.mouse_move_states.push(state.clone());
                }
                VirtualBinding::MouseButton(button, _) => {
                    self.mouse_button_states
                        .entry(*button)
                        .or_default()
                        .push(state.clone());
                }
                VirtualBinding::MouseWheelX(_) | VirtualBinding::MouseWheelY(_) => {
                    self.mouse_wheel_states.push(state.clone());
                }
                VirtualBinding::Axis(axis, _, _) => {
                    self.axis_states
                        .entry(*axis)
                        .or_default()
                        .push(state.clone());
                }
            }
        }

        state
            .write()
            .expect("Poisoned lock for add bindings")
            .bindings
            .extend(new_bindings.iter().map(|b| VirtualBindingState {
                value: 0.0,
                binding: b.clone(),
            }));
    }

    pub(super) fn update_key_binding(&mut self, key: PhysicalKey, key_state: ElementState) {
        let states = self.states_by_key.get_mut(&key);

        if let Some(states) = states {
            for state in states {
                let mut state = state.write().expect("Poisoned lock for key update");
                state.update_from_key(key, key_state);
            }
        }
    }

    pub(super) fn update_mouse_move_binding(&mut self, delta: &glam::Vec2) {
        for state in &mut self.mouse_move_states {
            let mut state = state.write().expect("Poisoned lock for mouse move update");
            state.update_from_mouse(delta);
        }
    }

    pub(super) fn update_mouse_wheel_binding(&mut self, delta: &glam::Vec2) {
        for state in &mut self.mouse_wheel_states {
            let mut state = state.write().expect("Poisoned lock for mouse wheel update");
            state.update_from_mouse_wheel(delta);
        }
    }

    pub(super) fn update_mouse_button_binding(
        &mut self,
        button: MouseButton,
        button_state: ElementState,
    ) {
        let states = self.mouse_button_states.get_mut(&button);

        if let Some(states) = states {
            for state in states {
                let mut state = state
                    .write()
                    .expect("Poisoned lock for mouse button update");
                state.update_from_mouse_button(button, button_state);
            }
        }
    }

    pub(super) fn update_axis_binding(&mut self, axis: AxisId, axis_state: f32) {
        let states = self.axis_states.get_mut(&axis);

        if let Some(states) = states {
            for state in states {
                let mut state = state.write().expect("Poisoned lock for axis update");
                state.update_from_axis(axis, axis_state);
            }
        }
    }
}
