use std::{
    any::TypeId,
    error::Error,
    sync::{Arc, RwLock, RwLockReadGuard},
};

use vulkano::{device::Device, format::Format, memory::allocator::StandardMemoryAllocator};

pub trait Pipeline {
    fn load(
        &mut self,
        device: &Device,
        memory_allocator: &StandardMemoryAllocator,
        swapchain_format: Format,
        depth_format: Format,
    ) -> Result<(), Box<dyn Error>>;
}

pub trait Material {
    fn pipeline_type_id() -> TypeId
    where
        Self: Sized;

    fn load(
        &mut self,
        device: &Device,
        memory_allocator: &StandardMemoryAllocator,
    ) -> Result<(), Box<dyn Error>>;
}

pub enum MaterialState {
    Loading,
    Loaded,
}

pub enum MaterialError {
    PipelineNotFound,
}

pub struct MaterialManager {
    device: Arc<Device>,
    memory_allocator: Arc<StandardMemoryAllocator>,
    swapchain_format: Format,
    depth_format: Format,

    pipelines_id: Arc<RwLock<Vec<TypeId>>>,
    pipelines_state: Arc<RwLock<Vec<Arc<RwLock<MaterialState>>>>>,
    pipelines: Arc<RwLock<Vec<Arc<RwLock<dyn Pipeline>>>>>,

    materials_id: Arc<RwLock<Vec<TypeId>>>,
    materials_pipeline_id: Arc<RwLock<Vec<TypeId>>>,
    materials_pipeline: Arc<RwLock<Vec<Arc<RwLock<dyn Pipeline>>>>>,
    materials_pipeline_state: Arc<RwLock<Vec<Arc<RwLock<MaterialState>>>>>,
    materials_state: Arc<RwLock<Vec<Arc<RwLock<MaterialState>>>>>,
    materials: Arc<RwLock<Vec<Arc<RwLock<dyn Material>>>>>,
}

impl MaterialManager {
    pub fn new(
        device: Arc<Device>,
        memory_allocator: Arc<StandardMemoryAllocator>,
        swapchain_format: Format,
        depth_format: Format,
    ) -> Self {
        Self {
            device,
            memory_allocator,
            swapchain_format,
            depth_format,
            pipelines_id: Arc::new(RwLock::new(Vec::new())),
            pipelines_state: Arc::new(RwLock::new(Vec::new())),
            pipelines: Arc::new(RwLock::new(Vec::new())),
            materials_id: Arc::new(RwLock::new(Vec::new())),
            materials_pipeline_id: Arc::new(RwLock::new(Vec::new())),
            materials_pipeline: Arc::new(RwLock::new(Vec::new())),
            materials_pipeline_state: Arc::new(RwLock::new(Vec::new())),
            materials_state: Arc::new(RwLock::new(Vec::new())),
            materials: Arc::new(RwLock::new(Vec::new())),
        }
    }

    pub fn add_pipeline<P: Pipeline + Default + 'static>(&self) {
        let type_id = TypeId::of::<P>();
        let pipeline = Arc::new(RwLock::new(P::default()));

        let mut pipelines_id = self.pipelines_id.write().unwrap();
        let mut pipelines_state = self.pipelines_state.write().unwrap();
        let mut pipelines = self.pipelines.write().unwrap();

        pipelines_id.push(type_id);
        pipelines_state.push(Arc::new(RwLock::new(MaterialState::Loading)));
        pipelines.push(pipeline.clone());
    }

    pub fn add_material<M: Material + Default + 'static>(&self) -> Result<(), MaterialError> {
        let pipeline_id = M::pipeline_type_id();

        let pipeline_result = {
            let pipelines_id = self.pipelines_id.read().unwrap();
            let pipelines_state = self.pipelines_state.read().unwrap();
            let pipelines = self.pipelines.read().unwrap();

            pipelines_id
                .iter()
                .zip(pipelines.iter())
                .zip(pipelines_state.iter())
                .find(|((id, _), _)| *id == &pipeline_id)
                .map(|((_, pipeline), state)| (pipeline.clone(), state.clone()))
        };

        let (pipeline, pipeline_state) = match pipeline_result {
            Some(pipeline) => pipeline,
            None => {
                tracing::error!(
                    "Pipeline with id {pipeline_id:?} not found, please add it before adding a material"
                );
                return Err(MaterialError::PipelineNotFound);
            }
        };

        let type_id = TypeId::of::<M>();

        let mut materials_id = self.materials_id.write().unwrap();
        let mut materials_pipeline_id = self.materials_pipeline_id.write().unwrap();
        let mut materials_pipeline = self.materials_pipeline.write().unwrap();
        let mut materials_pipeline_state = self.materials_pipeline_state.write().unwrap();
        let mut materials_state = self.materials_state.write().unwrap();
        let mut materials = self.materials.write().unwrap();

        materials_id.push(type_id);
        materials_pipeline_id.push(pipeline_id);
        materials_pipeline.push(pipeline);
        materials_pipeline_state.push(pipeline_state);
        materials_state.push(Arc::new(RwLock::new(MaterialState::Loading)));
        materials.push(Arc::new(RwLock::new(M::default())));

        Ok(())
    }

    pub fn update_swapchain_format(&mut self, swapchain_format: Format) {
        if self.swapchain_format == swapchain_format {
            return;
        }

        self.swapchain_format = swapchain_format;
        self.mark_all_pipelines_as_loading();
    }

    fn mark_all_pipelines_as_loading(&self) {
        let pipelines_state = self.pipelines_state.write().unwrap();

        for state in pipelines_state.iter() {
            let mut state = state.write().unwrap();
            *state = MaterialState::Loading;
        }
    }

    fn load_pipelines(&self) {
        let pipelines_state = self.pipelines_state.read().unwrap();
        let pipelines = self.pipelines.read().unwrap();

        let iter = pipelines_state
            .iter()
            .zip(pipelines.iter())
            .filter(|(state, _)| {
                let state = state.read().unwrap();
                matches!(*state, MaterialState::Loading)
            });

        for (state, pipeline) in iter {
            let mut pipeline = pipeline.write().unwrap();
            let result = pipeline.load(
                &self.device,
                &self.memory_allocator,
                self.swapchain_format,
                self.depth_format,
            );

            match result {
                Ok(_) => {
                    let mut state = state.write().unwrap();
                    *state = MaterialState::Loaded;
                }
                Err(e) => {
                    tracing::error!("Failed to load pipeline: {e}");
                }
            }
        }
    }

    fn load_materials(&self) {
        let materials_state = self.materials_state.read().unwrap();
        let materials = self.materials.read().unwrap();

        let iter = materials_state
            .iter()
            .zip(materials.iter())
            .filter(|(state, _)| {
                let state = state.read().unwrap();
                matches!(*state, MaterialState::Loading)
            });

        for (state, material) in iter {
            let mut material = material.write().unwrap();
            let result = material.load(&self.device, &self.memory_allocator);

            match result {
                Ok(_) => {
                    let mut state = state.write().unwrap();
                    *state = MaterialState::Loaded;
                }
                Err(e) => {
                    tracing::error!("Failed to load material: {e}");
                }
            }
        }
    }

    fn render_materials<F>(&self, f: F)
    where
        F: Fn(RwLockReadGuard<'_, dyn Material>, RwLockReadGuard<'_, dyn Pipeline>),
    {
        let materials = self.materials.read().unwrap();
        let materials_state = self.materials_state.read().unwrap();
        let materials_pipeline = self.materials_pipeline.read().unwrap();
        let materials_pipeline_state = self.materials_pipeline_state.read().unwrap();

        materials
            .iter()
            .zip(materials_state.iter())
            .zip(materials_pipeline.iter())
            .zip(materials_pipeline_state.iter())
            .filter(|(((_, material_state), _), pipeline_state)| {
                let material_state = material_state.read().unwrap();
                let pipeline_state = pipeline_state.read().unwrap();
                matches!(*material_state, MaterialState::Loaded)
                    && matches!(*pipeline_state, MaterialState::Loaded)
            })
            .for_each(|(((material, _), pipeline), _)| {
                let material = material.read().unwrap();
                let pipeline = pipeline.read().unwrap();

                f(material, pipeline);
            });
    }
}
