use std::{f32::consts::FRAC_PI_2, sync::Arc};

use glam::{Mat4, Vec3, Vec4};
use vulkano::{
    Validated,
    buffer::{AllocateBufferError, Subbuffer},
    memory::allocator::StandardMemoryAllocator,
};

use crate::core::{input::InputManager, timer::Timer};

use super::mvp::Mvp;

// See docs/OPENGL_VULKAN_DIFF.md
const OPENGL_TO_VULKAN_Y_AXIS_FLIP: Mat4 = Mat4 {
    x_axis: Vec4::new(1.0, 0.0, 0.0, 0.0),
    y_axis: Vec4::new(0.0, -1.0, 0.0, 0.0),
    z_axis: Vec4::new(0.0, 0.0, 1.0, 0.0),
    w_axis: Vec4::new(0.0, 0.0, 0.0, 1.0),
};

#[derive(Default)]
pub struct Camera3D {
    projection: Mat4,

    position: Vec3,
    rotation: Vec3,
    aspect_ratio: f32,
    fov: f32,
    near: f32,
    far: f32,
}

impl Camera3D {
    pub fn new(aspect_ratio: f32, fov: f32, near: f32, far: f32) -> Self {
        Self {
            projection: Mat4::perspective_rh(fov, aspect_ratio, near, far),
            position: Vec3::ZERO,
            rotation: Vec3::ZERO,
            aspect_ratio,
            fov,
            near,
            far,
        }
    }

    pub fn update(
        &mut self,
        input_manager: &InputManager,
        timer: &Timer,
        movement_speed: f32,
        camera_sensitivity: f32,
        window_aspect_ratio: f32,
    ) {
        // Process camera rotation
        let camera_delta = camera_sensitivity * timer.delta_time();
        self.rotation += Vec3::new(
            (input_manager.get_virtual_input_state("mouse_y") * camera_delta).to_radians(),
            (input_manager.get_virtual_input_state("mouse_x") * camera_delta).to_radians(),
            0.0,
        );

        if self.rotation.x > FRAC_PI_2 {
            self.rotation = Vec3::new(FRAC_PI_2, self.rotation.y, 0.0);
        }

        if self.rotation.x < -FRAC_PI_2 {
            self.rotation = Vec3::new(-FRAC_PI_2, self.rotation.y, 0.0);
        }

        let movement_delta = movement_speed * timer.delta_time();

        let (yaw_sin, yaw_cos) = self.rotation.y.sin_cos();
        let forward = Vec3::new(yaw_cos, 0.0, yaw_sin).normalize();
        let right = Vec3::new(-yaw_sin, 0.0, yaw_cos).normalize();

        let tx = input_manager.get_virtual_input_state("move_right") * movement_delta;
        self.position += tx * right;

        let tz = input_manager.get_virtual_input_state("move_forward") * movement_delta;
        self.position += tz * forward;

        if self.aspect_ratio != window_aspect_ratio {
            self.aspect_ratio = window_aspect_ratio;
            self.projection =
                Mat4::perspective_rh(self.fov, self.aspect_ratio, self.near, self.far);
        }
    }

    pub fn set_projection(&mut self, projection: Mat4) {
        self.projection = projection;
    }

    pub fn get_position(&self) -> Vec3 {
        self.position
    }

    pub fn get_rotation(&self) -> Vec3 {
        self.rotation
    }

    pub fn create_buffer(
        &self,
        memory_allocator: &Arc<StandardMemoryAllocator>,
    ) -> Result<Subbuffer<[Mvp]>, Validated<AllocateBufferError>> {
        let (sin_pitch, cos_pitch) = self.rotation.x.sin_cos();
        let (sin_yaw, cos_yaw) = self.rotation.y.sin_cos();

        let view_matrix = Mat4::look_to_rh(
            self.position,
            Vec3::new(cos_pitch * cos_yaw, sin_pitch, cos_pitch * sin_yaw).normalize(),
            Vec3::Y,
        );

        Mvp {
            model: OPENGL_TO_VULKAN_Y_AXIS_FLIP.to_cols_array_2d(),
            view: view_matrix.to_cols_array_2d(),
            projection: self.projection.to_cols_array_2d(),
        }
        .into_buffer(memory_allocator)
    }
}
