use std::sync::Arc;
use vulkano::{
    command_buffer::{AutoCommandBufferBuilder, RenderingAttachmentInfo, RenderingInfo},
    image::view::ImageView,
    pipeline::graphics::viewport::Viewport,
    render_pass::{AttachmentLoadOp, AttachmentStoreOp},
};

/// Types de render passes disponibles
#[derive(Debug, Clone)]
pub enum RenderPassType {
    Standard,
    ShadowMap,
    PostProcess,
}

/// Configuration pour un render pass
#[derive(Debug, Clone)]
pub struct RenderPassConfig {
    pub pass_type: RenderPassType,
    pub clear_color: Option<[f32; 4]>,
    pub clear_depth: Option<f32>,
    pub load_op: AttachmentLoadOp,
    pub store_op: AttachmentStoreOp,
}

impl Default for RenderPassConfig {
    fn default() -> Self {
        Self {
            pass_type: RenderPassType::Standard,
            clear_color: Some([0.0, 0.0, 0.0, 1.0]),
            clear_depth: Some(1.0),
            load_op: AttachmentLoadOp::Clear,
            store_op: AttachmentStoreOp::Store,
        }
    }
}

/// Gestionnaire de render passes réutilisable
pub struct RenderPassManager;

impl RenderPassManager {
    /// Commence un render pass standard avec les paramètres donnés
    pub fn begin_standard_rendering(
        builder: &mut AutoCommandBufferBuilder<vulkano::command_buffer::PrimaryAutoCommandBuffer>,
        config: &RenderPassConfig,
        color_attachment: Arc<ImageView>,
        depth_attachment: Option<Arc<ImageView>>,
        window_size: [f32; 2],
    ) -> Result<(), Box<dyn std::error::Error>> {
        let viewport = Viewport {
            offset: [0.0, 0.0],
            extent: window_size,
            depth_range: 0.0..=1.0,
        };

        let mut rendering_info = RenderingInfo {
            color_attachments: vec![Some(RenderingAttachmentInfo {
                load_op: config.load_op,
                store_op: config.store_op,
                clear_value: config.clear_color.map(|c| c.into()),
                ..RenderingAttachmentInfo::image_view(color_attachment)
            })],
            depth_attachment: None,
            ..Default::default()
        };

        if let Some(depth_view) = depth_attachment {
            rendering_info.depth_attachment = Some(RenderingAttachmentInfo {
                load_op: AttachmentLoadOp::Clear,
                store_op: AttachmentStoreOp::DontCare,
                clear_value: config.clear_depth.map(|d| [d].into()),
                ..RenderingAttachmentInfo::image_view(depth_view)
            });
        }

        builder
            .begin_rendering(rendering_info)?
            .set_viewport(0, [viewport].into_iter().collect())?;

        Ok(())
    }

    /// Termine le render pass actuel
    pub fn end_rendering(
        builder: &mut AutoCommandBufferBuilder<vulkano::command_buffer::PrimaryAutoCommandBuffer>,
    ) -> Result<(), Box<dyn std::error::Error>> {
        builder.end_rendering()?;
        Ok(())
    }

    /// Crée une configuration pour un render pass shadow map
    pub fn shadow_map_config() -> RenderPassConfig {
        RenderPassConfig {
            pass_type: RenderPassType::ShadowMap,
            clear_color: None,
            clear_depth: Some(1.0),
            load_op: AttachmentLoadOp::Clear,
            store_op: AttachmentStoreOp::Store,
        }
    }

    /// Crée une configuration pour un render pass de post-processing
    pub fn post_process_config() -> RenderPassConfig {
        RenderPassConfig {
            pass_type: RenderPassType::PostProcess,
            clear_color: Some([0.0, 0.0, 0.0, 1.0]),
            clear_depth: None,
            load_op: AttachmentLoadOp::Load,
            store_op: AttachmentStoreOp::Store,
        }
    }
}
